/*
 * Created on 23.02.2004 by Joern Schumacher
 *
 */
package CH.ifa.draw.application;

import CH.ifa.draw.DrawPlugin;

import CH.ifa.draw.util.Command;
import CH.ifa.draw.util.CommandMenu;
import CH.ifa.draw.util.CommandMenuItem;

import java.awt.Component;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.JMenu;
import javax.swing.JMenuItem;


/**
 * This class manages the menus of the JHotDraw draw window.<br>
 * If a plugin wants to add a menu item, it uses the registerMenu() method.
 * A unique id needs to be calculated by the registering plugin to identify the registered item.
 * From there, two possibilities exist:
 * <ul>
 * <li>the plugin registers the item with the registerMenuItem(String, JMenuItem, String id)
 * method.</li>
 * <li>the plugin first calls setClientProperty(MenuManager.ID_PROPERTY, id) on the item
 * and then uses the registerMenuItem(String, JMenuItem) method to register it.</li>
 * <p>
 * Top level menus are identified by their label; if a menu item is added to
 * a top level menu that does not exist yet, it is created.
 * Separator items need not be unregistered: they are removed automatically if they appear
 * on the top or bottom of a menu if a item is removed.
 *
 * @author Joern Schumacher
 * @author Michael Duvigneau
 */
public class MenuManager {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                                             .getLogger(MenuManager.class);
    public static final String ID_PROPERTY = "ch.ifa.draw.menu.id";
    private static MenuManager instance;

    /**
     * A map from top level menu names to their menu objects.
     * Consists of pairs <code>(String, JMenu)</code>.
     **/
    private HashMap toplevelMenus;

    /**
     * A map from identification strings to their associated menu items.
     * Consists of pairs <code>(String, JMenuItem)</code>.
     **/
    private HashMap registeredMenus;

    /**
     * A list of top level menu names, representing the order of registered
     * top level menus.
     **/
    private List toplevelOrder = new Vector();

    /**
     * The currently running application.  If <code>null</code>, no
     * application is running.
     **/
    private DrawApplication application = null;

    /**
     * Indicates that the top level menu with the name
     * {@link DrawPlugin#WINDOWS_MENU} has been registered.
     * This menu has a special position at the end of the menu bar,
     * therefore all other menus (except the <code>Help</code> menu)
     * must be inserted in front of it.
     **/
    private boolean windowsMenuPresent = false;

    /**
     * Indicates that the top level menu with the name
     * {@link DrawPlugin#HELP_MENU} has been registered.
     * This menu has a special position at the end of the menu bar,
     * therefore all other menus must be inserted in front of it.
     **/
    private boolean helpMenuPresent = false;

    private MenuManager() {
        toplevelMenus = new HashMap();
        registeredMenus = new HashMap();


        // The windows menu is always present.
        // Because the application is not known yet, we just add the menu
        // to our preconfiguration tables.
        WindowsMenu winMenu = new WindowsMenu(DrawPlugin.WINDOWS_MENU);
        toplevelMenus.put(DrawPlugin.WINDOWS_MENU, winMenu);
        toplevelOrder.add(DrawPlugin.WINDOWS_MENU);
        windowsMenuPresent = true;
    }

    /**
     * Singleton access method
     * @return The MenuManager instance
     */
    public static MenuManager getInstance() {
        if (instance == null) {
            instance = new MenuManager();
        }
        return instance;
    }

    /**
     *
     * Register the given menu below the top level menu identified by the
     * parentMenu parameter.
     * IMPORTANT: The registered menu must have the client property
     * MenuManager.ID_PROPERTY set to a sensible ID (String).
     * (This is done via the setClientProperty() method).
     *
     * If you just want to add one menu item, we recommend using the
     * registerMenu(String parent, JMenuItem item, String id) below
     * since it set the client property itself and therefore is less error prone.
     *
     * @param parentMenu the top level menu in which to show the menu item.
     *                         If this menu does not yet exist, it will be created.
     * @param toRegister the menu item to add.
     */
    public synchronized void registerMenu(String parentMenu,
                                          JMenuItem toRegister) {
        // check if the parentMenu exists, else create it
        JMenu parent = (JMenu) toplevelMenus.get(parentMenu);
        String id = (String) toRegister.getClientProperty(ID_PROPERTY);
        if (id == null) {
            logger.error(getClass() + ": could not register menu " + toRegister
                         + ": please set " + ID_PROPERTY
                         + " with menu.setClientProperty().");
            return;
        }
        if (parent == null) {
            logger.debug(getClass() + ": creating new top level menu "
                         + parentMenu);

            int newPosition;
            if (parentMenu.equals(DrawPlugin.WINDOWS_MENU)) {
                throw new RuntimeException("Windows menu should be always present, but is not available!");
            } else if (parentMenu.equals(DrawPlugin.HELP_MENU)) {
                parent = new CommandMenu(parentMenu);
                newPosition = toplevelOrder.size();
                helpMenuPresent = true;
            } else {
                // we need a command menu so DrawApp will be able to check enabledness
                // and automatically trigger commands in CommandMenuItems
                parent = new CommandMenu(parentMenu);
                newPosition = toplevelOrder.size() - (helpMenuPresent ? 1 : 0)
                              - (windowsMenuPresent ? 1 : 0);
            }

            toplevelMenus.put(parentMenu, parent);
            toplevelOrder.add(newPosition, parentMenu);
            if (application != null) {
                application.addMenu(parent, newPosition);
            }
        }

        if (toRegister instanceof SeparatorItem) {
            // no need to unregister separators: they are removed automatically
            // registeredMenus.put(id, toRegister);
            parent.addSeparator();

        } else if (toRegister instanceof JMenuItem) {
            registeredMenus.put(id, toRegister);
            parent.add(toRegister);
        } else {
            logger.error(getClass() + ": cannot add " + toRegister
                         + " to menu " + parentMenu + ": class "
                         + toRegister.getClass() + " invalid.");
            removeMenuIfEmpty(parentMenu, parent);
        }
        logger.debug(getClass() + ": the menu " + toRegister.getText()
                     + " has been registered under " + parentMenu + ", id "
                     + toRegister.getClientProperty(ID_PROPERTY));
    }

    /**
     * Register the given menu item, setting the client property
     * to the given id string.
     *
     * @param parentMenu
     * @param toRegister
     * @param id
     */
    public void registerMenu(String parentMenu, JMenuItem toRegister, String id) {
        toRegister.putClientProperty(ID_PROPERTY, id);
        registerMenu(parentMenu, toRegister);
    }

    /**
     * Unregister a previously registered menu.
     * Also check out the unregisterMenu(SeparatorItem) method
     * to avoid code smell.
     *
     * @param item
     */
    public void unregisterMenu(JMenuItem item) {
        if (item instanceof SeparatorItem) {
            logger.debug(getClass() + ": no need to unregister separator.");
            return;
        }
        String id = (String) item.getClientProperty(ID_PROPERTY);
        if (id == null) {
            logger.debug(getClass() + " unregistering item " + item.getText()
                         + ": no client property set!");
            return;
        }
        unregisterMenu(id);
    }

    public synchronized void unregisterMenu(String id) {
        // this should work...
        // JMenu toplevel = (JMenu) item.getParent();
        // it didnt: a JPopupMenu appeared from nowhere.
        // i'll have to look for the parent manually
        JMenu toplevel = null;
        Iterator tlMenus = toplevelMenus.values().iterator();
        JMenuItem item = null;
        while (toplevel == null && tlMenus.hasNext()) {
            JMenu c = (JMenu) tlMenus.next();
            String tlLabel = c.getText();

            Component[] comp = c.getMenuComponents();
            for (int i = 0; i < comp.length; i++) {
                if (comp[i] instanceof JMenuItem) {
                    JMenuItem currentItem = (JMenuItem) comp[i];
                    String currentId = (String) currentItem.getClientProperty(ID_PROPERTY);
                    if (id.equals(currentId)) {
                        toplevel = (JMenu) c;
                        item = currentItem;
                        break;
                    }
                }
            }
        }
        if (item == null) {
            logger.debug(getClass() + ".unregisterMenu(): menu " + id
                         + " not found!");
            return;
        }

        // now, wasn't that fun?
        logger.debug(getClass() + ": unregistering menu " + item.getText()
                     + " from " + toplevel.getText());

        toplevel.remove(item);


        // if the first item in the menu is now a separator,
        // remove it.
        while (toplevel.getItemCount() > 0 && toplevel.getItem(0) == null) {
            logger.debug(getClass() + ": removing top separator.");
            toplevel.remove(0);
        }

        // same for the last one.
        while (toplevel.getItemCount() > 0
                       && toplevel.getItem(toplevel.getItemCount() - 1) == null) {
            logger.debug(getClass() + ": removing top separator.");
            toplevel.remove(toplevel.getItemCount() - 1);
        }

        removeMenuIfEmpty(toplevel.getText(), toplevel);
        registeredMenus.remove(item.getClientProperty(ID_PROPERTY));
    }

    /**
     * Describe <code>removeMenuIfEmpty</code> method here.
     *
     * @param parentMenu a <code>String</code> value
     * @param menuObject a <code>JMenu</code> value
     **/
    private void removeMenuIfEmpty(String parentMenu, JMenu menuObject) {
        // Never remove the windows menu.
        if (DrawPlugin.WINDOWS_MENU.equals(parentMenu)) {
            return;
        }

        // check if menuObject menu has become empty
        if (menuObject.getMenuComponentCount() == 0) {
            logger.debug(getClass() + ": the last MenuItem of menu "
                         + parentMenu + " has been removed, removing menu");
            if (application != null) {
                application.removeMenu(menuObject);
            }
            toplevelMenus.remove(parentMenu);
            toplevelOrder.remove(parentMenu);
            if (DrawPlugin.HELP_MENU.equals(parentMenu)) {
                helpMenuPresent = false;
            }
        }
    }

    /**
     * This method should be called if the gui is opened (with the new instance as parameter)
     * or closed (with null to allow garbage collection).
     *
     * @param app the Gui object that has just been opened or null if a previous
     * instance is closed.
     */
    void setGui(DrawApplication app) {
        application = app;

        if (application != null) {
            logger.debug(getClass() + ": adding menus to application object "
                         + app);


            // create all those registered menus
            // applying the order in which they were created
            Iterator labels = toplevelOrder.iterator();
            while (labels.hasNext()) {
                String label = (String) labels.next();
                JMenu add = (JMenu) toplevelMenus.get(label);
                application.addMenu(add);
            }
        }
    }

    /**
     * Convenience method to be able to create a
     * separator JMenuItem on the fly.
     * This method does NOT assure that the id is unique
     * so the client has to assure it.
     * If you want the framework to handle unique ids,
     * use a SeparatorFactory object.
     * @param id  the id for referencing that is set as client property.
     * @return a new separator menu item with the given <code>id</code>.
     */
    public static JMenuItem createSeparator(String id) {
        return new SeparatorItem(id);
    }

    JMenu getToplevelMenu(String name) {
        return (JMenu) toplevelMenus.get(name);
    }

    /**
     * Returns the special windows menu.
     * Although this menu is a toplevel menu, it cannot be unregistered.
     * Currently it is always created with the <code>MenuManager</code>.
     *
     * @return the registered <code>WindowsMenu</code> instance.
     **/
    public WindowsMenu getWindowsMenu() {
        return (WindowsMenu) getToplevelMenu(DrawPlugin.WINDOWS_MENU);
    }

    /**
     * Convenience method to create a MenuItem from a given Command
     * that will be executed when the item is chosen.
     *
     * @param command   the command to execute when this menu item is
     *                  activated.
     * @param shortcut  the AWT keyboard shortcut to assign to this menu entry.
     * @return  the created menu item.
     */
    public static CommandMenuItem createMenuItem(Command command, int shortcut) {
        CommandMenuItem item = new CommandMenuItem(command, shortcut);
        return item;
    }

    // an evil tag class to be able to add separators to menus...
    // can't think of any other way
    private static class SeparatorItem extends JMenuItem {
        private int position = -1;
        private JMenu parent;

        public SeparatorItem(String label) {
            // we need a label to be able to remove the menu again
            super(label);
        }

        public void setPosition(int index) {
            position = index;
        }

        public int getPosition() {
            return position;
        }

        public void setParentMenu(JMenu parent) {
            this.parent = parent;
        }

        public JMenu getParentMenu() {
            return parent;
        }
    }

    /**
     * Use this class to create Separators in your menus.
     * <br>
     * In the method where you create your menu, instantiate a
     * SeparatorFactory with your Plugin's ID
     * (the one you use in your putClientProperty(MenuManager.ID_PROPERTY, ...)-call).
     * The factory will then create JMenuItems that will be identified by the MenuManager,
     * so it will add a separator if you call menuManager.registerMenu(String, JMenuItem).
     * <br>
     * The factory will set the proper ids, so you don't have to worry about that.
     * Actually, you shouldn't.
     *
     * @author Joern Schumacher
     *
     */
    public static class SeparatorFactory {
        private int count;
        private String id;

        public SeparatorFactory(String id) {
            this.id = id;
        }

        public JMenuItem createSeparator() {
            JMenuItem item = new SeparatorItem(id + ".sep" + count);
            item.putClientProperty(ID_PROPERTY, id + ".sep" + count++);
            return item;
        }
    }
}