package de.renew.plugin;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.SimpleLayout;
import org.apache.log4j.xml.DOMConfigurator;

import de.renew.plugin.DependencyCheckList.DependencyElement;
import de.renew.plugin.command.CLCommand;
import de.renew.plugin.command.ExitCommand;
import de.renew.plugin.command.GCCommand;
import de.renew.plugin.command.GetPropertyCommand;
import de.renew.plugin.command.InfoCommand;
import de.renew.plugin.command.ListCommand;
import de.renew.plugin.command.LoadCommand;
import de.renew.plugin.command.NoOpCommand;
import de.renew.plugin.command.ScriptCommand;
import de.renew.plugin.command.SetPropertyCommand;
import de.renew.plugin.command.UnloadCommand;
import de.renew.plugin.load.PluginLoaderComposition;
import de.renew.plugin.load.SimplePluginLoader;
import de.renew.plugin.locate.PluginJarLocationFinder;
import de.renew.plugin.locate.PluginLocationFinders;
import de.renew.plugin.locate.PluginSubDirFinder;

import java.io.File;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.Serializable;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;


/**
 * This class is the central management facility for providing
 * Plugins with access to Renew.
 * When the Singleton instance of this class is created, the statical plugins will be loaded.
 * Furthermore, this class will be the interface for
 * dynamical plugins that can log into it at run-time.
 */
public class PluginManager implements Serializable {
    public static Logger logger = Logger.getLogger(PluginManager.class);

    /**
    * The timeout for cleaning up a plugin, in microseconds.
    * The value is fixed to {@value} microseconds.
    **/
    public static final int CLEANUP_TIMEOUT = 20000;

    /**
     * The name of the system property, which define the path for the plugins.
     */
    public static final String PLUGIN_LOCATIONS_PROPERTY = "pluginLocations";

    /**
     * The relative path to the property file with the
     * the log4j configuration. The path is used
     * relative to the directory where the loader.jar
     * is located.
     */
    public static final String LOG4J_PROPERTIES_FILE = "log4j.properties";

    /**
     * The relative path to the default xml file with the
     * the log4j configuration. The path is used
     * relative to the directory where the loader.jar
     * is located.
     */
    public static final String LOG4J_XML_FILE = "log4j.xml";

    /**
     * The name of the property to set the directory
     * for the log files. This property key can be used
     * in the log4j configuration file log4j.properties.
     * See the default log4j.properties file for an
     * example how to use the property logs.home.
     */
    public static final String LOG_HOME_PROPERTY_NAME = "logs.home";

    /**
     * The name of the property to set the location of the
     * configuration file for the configuration of the
     * logging framework.
     */
    public static final String LOG_CONFIG_PROPERTY_NAME = "log4j.configuration";

    /**
     * The relative path to the default directory
     * where the log files should be located.
     * The path is relative to the renew distribution
     * directory where the loader.jar is located.
     */
    public static final String LOG_HOME_DEFAULT_LOCATION = "renewlogs";

    /** Singleton member instance. **/
    protected static PluginManager _instance;

    /** The list managing the dependencies among the plugins. **/
    DependencyCheckList _dependencyList = new DependencyCheckList();

    /**
     * The map of recognized commands and their associated executors.
     * Maps from <code>String</code> to {@link CLCommand} objects.
     **/
    private Map<String, CLCommand> _commands = Collections.synchronizedMap(new TreeMap<String, CLCommand>());

    /**
     * Contains the listeners that register for PluginManager events.
     */
    private Set<IPluginManagerListener> _managerListener;

    /**
     * The set of plugins preventing the system from automatical
     * termination.
     **/
    private Set<IPlugin> _blockers = Collections.synchronizedSet(new HashSet<IPlugin>());

    /**
     * This flag indicates that the system termination thread is
     * running. It is set by the {@link #stopSynchronized()}
     * method only.
     **/
    private boolean _Terminating = false;
    protected PluginLocationFinders _locationFinder = PluginLocationFinders
                                                                                                                                              .getInstance();
    protected PluginLoaderComposition _loader = new PluginLoaderComposition();
    private PluginClassLoader _pluginClassLoader;
    private ClassLoader _bottomClassLoader;
    private ClassLoader _systemClassLoader;

    /**
     * The location of loader.jar
     */
    private static URL _loaderLocation;

    /**
     * Get the location of loader.jar
     * @return url of loader.jar
     */
    public static URL getLoaderLocation() {
        if (_loaderLocation == null) { // to prevent NullPointerException
            URL url = Loader.class.getProtectionDomain().getCodeSource()
                                  .getLocation();
            String base = url.toExternalForm();
            base = base.substring(0, base.lastIndexOf("/"));
            try {
                _loaderLocation = new URL(base + "/");
            } catch (MalformedURLException e) {
                e.printStackTrace();
            }
        }
        return _loaderLocation;
    }

    protected PluginManager() {
        configureLogging();

        _managerListener = new HashSet<IPluginManagerListener>();

        // create the PluginClassLoader
        //_pluginClassLoader = new PluginClassLoader(getClass().getClassLoader());
        _pluginClassLoader = new PluginClassLoader(getClass().getClassLoader());

        // get the SystemClassLoader (view of Plugin System)
        _systemClassLoader = this.getClass().getClassLoader();
        //ClassLoader.getSystemClassLoader(); //// old
        // create the default BottomClassLoader
        _bottomClassLoader = new BottomClassLoader(_pluginClassLoader);

        initLocationFinders();

        // initialize the loaders
        _loader = new PluginLoaderComposition();
        _loader.addLoader(new SimplePluginLoader());

        _commands.put("", new NoOpCommand());
        _commands.put("help", new HelpCommand());
        _commands.put("exit", new ExitCommand());
        _commands.put("load", new LoadCommand());
        _commands.put("list", new ListCommand());
        _commands.put("unload", new UnloadCommand());
        _commands.put("info", new InfoCommand());
        _commands.put("script", new ScriptCommand());
        _commands.put("set", new SetPropertyCommand());
        _commands.put("get", new GetPropertyCommand());
        _commands.put("gc", new GCCommand());
        _commands.put("packageCount",
                      _pluginClassLoader.new PackageCountCommand());
    }

    public static synchronized PluginManager getInstance() {
        if (_instance == null) {
            _instance = new PluginManager();
            _instance.initPlugins();
        }

        return _instance;
    }

    /**
     * Adds a Listener to the PluginManager
     * @param l the listener
     */
    public void addPluginManagerListener(IPluginManagerListener l) {
        _managerListener.add(l);
    }

    /**
     * Removes a Listener from the PluginManager
     * @param l the listener
     */
    public void removePluginManagerListener(IPluginManagerListener l) {
        _managerListener.remove(l);
    }

    private void serviceAdded(Collection<String> services) {
        for (IPluginManagerListener listener : _managerListener) {
            for (String service : services) {
                listener.serviceAdded(service);
            }
        }
    }

    private void serviceRemoved(Collection<String> services) {
        for (IPluginManagerListener listener : _managerListener) {
            for (String service : services) {
                listener.serviceRemoved(service);
            }
        }
    }

    public void addPlugin(IPlugin plugin)
            throws DependencyNotFulfilledException {
        if (!checkDependenciesFulfilled(plugin)) {
            throw new DependencyNotFulfilledException("Cannot add " + plugin);
        }

        try {
            logger.debug("************ INITIALIZING " + plugin.getName()
                         + " ********");
            plugin.init();
            _dependencyList.addElement(new DependencyElement(plugin));
            serviceAdded(plugin.getProperties().getProvisions());
        } catch (RuntimeException e) {
            logger.error("PluginManager: adding of Plugin " + plugin
                         + " failed: " + e);
            logger.debug(e.toString(), e);
        } catch (LinkageError e) {
            logger.error("PluginManager: adding of Plugin " + plugin
                         + " failed: " + e);
            logger.debug(e.toString(), e);
        }
    }

    /**
     * Checks whether the dependencies specified by the given
     * <code>PluginProperties</code> object are fulfilled by the current
     * set of loaded plugins.
     *
     * @param props  the properties that specify the dependencies.
     *
     * @return <code>true</code>, if a plugin with the given <code>props</code>
     *         could be added to the set of plugins. Returns <code>false</code>,
     *         if some dependency of the plugin would not be fulfilled.
     **/
    public synchronized boolean checkDependenciesFulfilled(PluginProperties props) {
        return _dependencyList.dependencyFulfilled(new DependencyElement(props));
    }

    /**
     * Checks whether the dependencies specified by the given
     * <code>IPlugin</code> object are fulfilled by the current
     * set of loaded plugins.
     *
     * @param plugin  the plugin that specifies the dependencies.
     *
     * @return <code>true</code>, if the given <code>plugin</code> could
     *         be added to the set of plugins. Returns <code>false</code>,
     *         if some dependency of the plugin would not be fulfilled.
     **/
    public synchronized boolean checkDependenciesFulfilled(IPlugin plugin) {
        return _dependencyList.dependencyFulfilled(new DependencyElement(plugin));
    }

    /**
     * Returns the plugin with the given name, null if
     * no such plugin is present.
     */
    public IPlugin getPluginByName(String pluginName) {
        IPlugin found = null;
        logger.debug("PluginManager looking for " + pluginName);
        List plugins = getPlugins();
        for (int i = 0; i < plugins.size(); i++) {
            IPlugin current = (IPlugin) plugins.get(i);
            if (current.getName().equals(pluginName)) {
                found = current;
                break;
            }
        }
        if (found == null) {
            found = getPluginByAlias(pluginName);
        }
        return found;
    }

    /**
     * Returns the plugin with the given alias, null if
     * no such plugin is present.
     */
    public IPlugin getPluginByAlias(String pluginAlias) {
        IPlugin found = null;
        logger.debug("PluginManager looking for " + pluginAlias);
        List<Object> plugins = getPlugins();
        for (int i = 0; i < plugins.size(); i++) {
            IPlugin current = (IPlugin) plugins.get(i);

            if (pluginAlias.equals(current.getAlias())) {
                found = current;
                break;
            }
        }
        return found;
    }

    /**
     * Returns a Collection containing the Plugins that provide
     * the given service.
     */
    public Collection<IPlugin> getPluginsProviding(String service) {
        Collection<IPlugin> result = new Vector<IPlugin>();
        Iterator it = getPlugins().iterator();
        while (it.hasNext()) {
            IPlugin p = (IPlugin) it.next();
            if (p.getProperties().getProvisions().contains(service)) {
                result.add(p);
            }
        }
        return result;
    }

    /**
     * Iterate through all plugins and initialize them.
     */
    public void initPlugins() {
        Iterator plugins = getPlugins().iterator();

        while (plugins.hasNext()) {
            IPlugin p = (IPlugin) plugins.next();
            logger.debug("initializing " + p.getName() + ", loaded from "
                         + p.getClass().getClassLoader());
            logger.debug("************ INITIALIZING " + p.getName()
                         + " ********");
            p.init();
            serviceAdded(p.getProperties().getProvisions());
        }
    }

    /**
     * Return the ClassLoader instance used to load the plugins.
     */
    public PluginClassLoader getPluginClassLoader() {
        return _pluginClassLoader;
    }

    /**
     * Returns the ClassLoader instance used to load the user defined .
     */
    public ClassLoader getBottomClassLoader() {
        return _bottomClassLoader;
    }

    /**
     * Returns the system classloader
     */
    public ClassLoader getSystemClassLoader() {
        return _systemClassLoader;
    }

    /**
     * Get the PluginLoader used to create plugin instances
     */
    PluginLoaderComposition getPluginLoader() {
        return _loader;
    }

    /**
     * Finds all candidate locations and tries to load a plugin from each
     * location. The currently configured <code>PluginLoader</code> is
     * used for the job.
     **/
    public synchronized void loadPlugins() {
        Collection all = _loader.loadPlugins();
    }

    /**
     * Loads a plugin from the given URL, if possible. The currently
     * configured <code>PluginLoader</code> is used for the job.
     *
     * @param url  the plugin's location as <code>URL</code>.
     **/
    public synchronized IPlugin loadPlugin(URL url) {
        return _loader.loadPluginFromURL(url);
    }

    /**
     * Return a list of all loaded Plugins.
     */
    public List<Object> getPlugins() {
        return _dependencyList.getFulfilledObjects();
    }

    /**
     * Register the given command. It will be called when the user
     * types the given String into the console.
     * It is not checked whether the String already exists,
     * so an existing command can be overridden!
     *
     * @param command The String identifying
     * @param clcommand The command to add
     */
    public void addCLCommand(String command, CLCommand clcommand) {
        _commands.put(command, clcommand);
    }

    /**
     * Unregister the command identified by the given String.
     */
    public void removeCLCommand(String command) {
        _commands.remove(command);
    }

    /**
     * Return a Map containing all commands as values,
     * with the identifying Strings as keys.
     */
    public Map<String, CLCommand> getCLCommands() {
        return _commands;
    }

    /**
     * Adds the given plugin to the set of exit blockers.
     * If the exit blocker set becomes non-empty, the plugin system will
     * not be automatically terminated.
     *
     * @param blocker the plugin that prevents automatic termination of the
     *                plugin system.
     **/
    public void blockExit(IPlugin blocker) {
        if (blocker != null) {
            logger.debug("PluginManager: registering exit blocker " + blocker);
            _blockers.add(blocker);
        }
    }

    /**
     * Removes the given plugin from the set of exit blockers.
     * If the exit blocker set becomes empty, the plugin system is
     * automatically terminated.
     *
     * @param blocker the plugin that should not any longer prevent
     *                automatic termination of the plugin system.
     *                If this plugin object is not included in the set of
     *                exit blockers, the set will not be modified.
     *                Nevertheless, the set will be checked for emptyness.
     **/
    public void exitOk(IPlugin blocker) {
        synchronized (_blockers) {
            if (blocker != null) {
                logger.debug("PluginManager: unregistering exit blocker "
                             + blocker);
                _blockers.remove(blocker);
            }
            checkExit();
        }
    }

    /**
     * Checks whether there are any exit blockers registered, and shuts
     * down the plugin system, if not.
     **/
    public boolean checkExit() {
        synchronized (_blockers) {
            if (_blockers.isEmpty()) {
                logger.debug("PluginManager: no active plugins, shutting down.");

                // Since there might still be AWT threads running,
                // shut down the Java VM manually.
                // The asynchronous stop method is used to decouple the system
                // termination from the plugin that unknowingly initiated the
                // process. This also unrolls the endless loop when a plugin on
                // termination deregisters itself as exit blocker.
                stop();
                return true;
            } else if (logger.isDebugEnabled()) {
                logger.debug("Active plugins blocking exit: "
                             + CollectionLister.toString(_blockers));
            }
        }
        return false;
    }

    /**
     * Terminates all plugins and the whole Java system. Whenever the
     * termination of any plugin fails, the termination process is
     * cancelled, too.
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * Plugins are then terminated in sequence (with respect to plugin
     * dependencies), but concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     **/
    public void stop() {
        logger.debug("Initiating PluginManager termination.");
        Thread terminationThread = new Thread() {
            public void run() {
                if (stopSynchronized()) {
                    System.exit(0);
                }
            }
        };
        terminationThread.start();
    }

    /**
     * Stops the given plugin.
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * The plugin is then terminated (with respect to plugin dependencies)
     * concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     *
     * @param p the plugin to remove from the plugin system.
     **/
    public void stop(final IPlugin p) {
        logger.debug("Initiating termination of " + p + ".");
        Thread terminationThread = new Thread() {
            public void run() {
                stopSynchronized(p);
            }
        };
        terminationThread.start();
    }

    /**
     * Stops the given plugins. Works through the list <em>from back to
     * front</em>. If any plugin cannot be stopped due to some other dependent
     * plugin (it does not matter whether it is or is not included in the
     * list in front of the plugin it depends on), the list is not further
     * processed.
     * <p>
     * Example: Assume that plugin B depends on plugin A.
     * A call with the list <code>[A, B]</code> will succeed.
     * But a call with the list <code>[B, A]</code> will fail because
     * plugin A cannot be terminated due to the dependent plugin B.
     * </p>
     * <p>
     * The reason for the backward processing is that lists of recursively
     * dependent plugins are easier to generate with the most basic plugin
     * at first position.
     * </p>
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * The plugins are then terminated (with respect to plugin dependencies)
     * concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     *
     * @param plugins the plugins to remove from the plugin system.
     **/
    public void stop(final List plugins) {
        logger.debug("Initiating termination of " + plugins + ".");
        Thread terminationThread = new Thread() {
            public void run() {
                stopSynchronized(plugins);
            }
        };
        terminationThread.start();
    }

    /**
     * Tells whether the plugin system is about to terminate.
     * This method only returns true, if the {@link #stop()}
     * method has been used to trigger the system termination.
     * <p>
     * Please be aware that this method's result may be out of
     * date before you are able to interpret it. It is intended
     * to be used by terminated plugins that want to know
     * whether it's just them or the whole system...
     * </p>
     *
     * @return <code>true</code> if the concurrent system
     *         termination thread is running.
     **/
    public boolean isStopping() {
        return _Terminating;
    }

    /**
     * Terminates all plugins.
     * This method is called internally by {@link #stop()}.
     **/
    public synchronized boolean stopSynchronized() {
        logger.debug("Stopping plugin system asynchronously.");
        _Terminating = true;
        try {
            List pls = _dependencyList.getFulfilledObjects();
            return stopSynchronized(pls);
        } finally {
            _Terminating = false;
        }
    }

    /**
     * Terminates all plugins in the given list.
     * This method is called internally by {@link #stopSynchronized()}
     * and {@link #stop(List)}.
     **/
    private synchronized boolean stopSynchronized(List pls) {
        int limit = pls.size();
        for (int i = pls.size() - 1; i >= 0; i--) {
            // ask all plugins if they can be stopped.
            final IPlugin toStop = (IPlugin) pls.get(i);
            Object sync = new Object();
            if (logger.isDebugEnabled()) {
                logger.debug("Preparing stop of " + toStop);
            }
            int waitSeconds = toStop.getProperties()
                                    .getIntProperty("cleanupTimeout",
                                                    CLEANUP_TIMEOUT);
            int firstWait = waitSeconds;
            int secondWait = 1;
            if (waitSeconds > 10) {
                firstWait = waitSeconds / 10;
                secondWait = waitSeconds - firstWait;
            }
            if (logger.isTraceEnabled()) {
                logger.trace("canShutDown timeout is " + waitSeconds
                             + " milliseconds (" + firstWait + "/" + secondWait
                             + ").");
            }
            SynchronizedThread ct = new SynchronizedThread(new Command() {
                    public boolean execute() {
                        return toStop.canShutDown();
                    }
                }, sync);
            try {
                ct.start();
                if (logger.isTraceEnabled()) {
                    logger.trace("Started canShutDown thread, synching on "
                                 + sync);
                }
                synchronized (sync) {
                    // has thread already been finished?
                    if (ct.didFinish()) {
                        // do nothing
                    } else if (waitSeconds == -1) {
                        // stay a while...
                        // stay... forever!
                        if (logger.isTraceEnabled()) {
                            logger.trace("Waiting (unlimited) for canShutDown thread to finish, synched on "
                                         + sync);
                        }
                        sync.wait();
                    } else {
                        if (logger.isTraceEnabled()) {
                            logger.trace("Waiting (first part) for canShutDown thread to finish, synched on "
                                         + sync);
                        }
                        sync.wait(firstWait);
                        if (!ct.didFinish()) {
                            logger.info("Waiting for " + toStop
                                        + " to confirm termination...");
                            sync.wait(secondWait);
                        }
                        if (logger.isTraceEnabled()) {
                            logger.trace("Done waiting, canShutDown thread "
                                         + (ct.didFinish() ? "finished"
                                                           : "did not finish")
                                         + ", synched on " + sync);
                        }
                    }
                }
            } catch (InterruptedException e) {
                if (logger.isTraceEnabled()) {
                    logger.trace("Interrupted timeout waiting on " + sync);
                }
            }
            if (!ct.hadSuccess()) {
                logger.warn(toStop + " did not confirm termination in time.");
                return false;
            } else if (!ct.hadSuccess()) {
                logger.warn(toStop + " said it cannot shut down.");
                return false;
            } else {
                logger.debug(toStop + " said it can shut down.");
            }
        }
        for (int i = pls.size() - 1; i >= 0; i--) {
            IPlugin toStop = (IPlugin) pls.get(i);
            try {
                _dependencyList.removeElement(toStop);
            } catch (DependencyNotFulfilledException e) {
                logger.error(e.getMessage());
                logger.error("list of dependent plugins:");
                logger.error(CollectionLister.toString(e.getElements()));
                return false;
            }

            if (!stopSynchronized(toStop)) {
                logger.warn("stop cancelled by plugin " + toStop);
                _dependencyList.addElement(new DependencyElement(toStop));
                return false;
            }
        }
        return true;
    }

    /**
     * Stops the given plugin.
     * This method is called internally by {@link #stop(IPlugin)} and
     * {@link #stopSynchronized(List)}.
     **/
    private synchronized boolean stopSynchronized(final IPlugin p) {
        try {
            synchronized (_dependencyList) {
                _dependencyList.removeElement(p);
            }
        } catch (DependencyNotFulfilledException e) {
            logger.error(e.getMessage());
            logger.error("list of dependent plugins:");
            logger.error(CollectionLister.toString(e.getElements()));
            return false;
        }

        logger.debug("stopping " + p);
        boolean result;
        Object sync = new Object();
        int waitSeconds = p.getProperties()
                           .getIntProperty("cleanupTimeout", CLEANUP_TIMEOUT);
        if (logger.isTraceEnabled()) {
            logger.trace("Cleanup timeout is " + waitSeconds + " milliseconds.");
        }
        SynchronizedThread ct = new SynchronizedThread(new Command() {
                public boolean execute() {
                    return p.cleanup();
                }
            }, sync);


        // prevent stop from being called by the manager because
        // it just shut down the last plugin with an exit block.
        IPlugin blockLock = new PluginAdapter(PluginProperties.getUserProperties());
        _blockers.add(blockLock);
        try {
            ct.start();
            if (logger.isTraceEnabled()) {
                logger.trace("Started cleanup thread, synching on " + sync);
            }
            synchronized (sync) {
                if (!ct.didFinish()) {
                    // stay a while...
                    if (waitSeconds == -1) {
                        // stay... forever!
                        if (logger.isTraceEnabled()) {
                            logger.trace("Waiting (unlimited) for cleanup thread to finish, synched on "
                                         + sync);
                        }
                        sync.wait();
                    } else {
                        if (logger.isTraceEnabled()) {
                            logger.trace("Waiting for cleanup thread to finish, synched on "
                                         + sync);
                        }
                        sync.wait(waitSeconds);
                    }
                }
                if (logger.isTraceEnabled()) {
                    logger.trace("Done waiting, cleanup thread "
                                 + (ct.didFinish() ? "finished" : "did not finish")
                                 + ", synched on " + sync);
                }
            }
        } catch (InterruptedException e) {
            if (logger.isTraceEnabled()) {
                logger.trace("Interrupted timeout waiting on " + sync);
            }
        }
        _blockers.remove(blockLock);
        logger.debug(p + " stopped");
        result = ct.hadSuccess();
        if (!result) {
            synchronized (_dependencyList) {
                _dependencyList.addElement(new DependencyElement(p));
            }
        } else {
            //success:
            serviceRemoved(p.getProperties().getProvisions());
        }
        return result;
    }

    /**
     * Start the Renew Plugin System
     * @param args the command line parameters
     * @param url the url of the Loader.class needs a trailing slash
     */
    public static void main(String[] args, URL url) {
        _loaderLocation = url;
        main(args);
    }

    /**
     * Start the Renew Plugin System.
     * Call {@link #main(String[], URL)} instead of this method.
     */
    public static void main(String[] args) {
        getLoaderLocation();

        Thread.currentThread()
              .setContextClassLoader(PluginManager.getInstance()
                                                  .getBottomClassLoader());
        CLCommand command = null;
        PluginManager pm = getInstance();
        pm.loadPlugins();
        Iterator plugins = pm.getPlugins().iterator();
        while (plugins.hasNext()) {
            logger.info("loaded plugin: "
                        + ((IPlugin) plugins.next()).getName());
        }
        if (args.length > 0) {
            String[] rest = new String[args.length - 1];
            for (int i = 0; i < args.length - 1; i++) {
                rest[i] = args[i + 1];
            }
            command = pm._commands.get(args[0]);
            if (command != null) {
                if (logger.isDebugEnabled()) {
                    StringBuffer fullCommandLine = new StringBuffer();
                    fullCommandLine.append(args[0]);
                    for (int i = 0; i < rest.length; i++) {
                        fullCommandLine.append(' ');
                        fullCommandLine.append(rest[i]);
                    }
                    logger.debug("Executing initial command: "
                                 + fullCommandLine.toString());
                }

                // TODO: Instead of System.out we should use a print stream
                //       that maps to the logger.
                command.execute(rest, System.out);
            } else {
                logger.warn("Unknown initial command (ignored): " + args[0]);
            }
        }
        pm.checkExit();
    }

    /**
     * Configures the log4j framework by trying to load properties from a
     * configuration file in the config directory of the renew distribution
     * home directory. To configure the log4j framework this method first
     * look if the property log4j.configuration is set. If this property
     * is not set, an existing xml configuration file with the name log4j.xml
     * located in the config directory of renew will be used. If no such
     * file exists, a property configuration file with the name log4j.properties
     * will be used.
     * <p>
     * If there are configuration files with the upper names <em>and an
     * initial dot</em> in the user home directory, first the default files
     * from the config directory will be read and than additionally the
     * config files from the users home directory.
     * The properties set in the users config files overwrite the default
     * properties.
     * </p>
     * <p>
     * If the system property "log4j.configuration" is set only the propery
     * settings from the given file are used.
     * </p>
     * TODO: Currenty only the XML users configuration file is used to configure
     *       the logging framework if one exists, because the semantic merging
     *       of two xml configuration files seems a little strange.
     */
    public static void configureLogging() {
        boolean configurationDone = false;
        Vector<String> logMessages = new Vector<String>();
        String logConfigSource = null;

        try {
            String configLocation = System.getProperty(LOG_CONFIG_PROPERTY_NAME);

            if (configLocation != null) {
                if (configLocation.endsWith(".xml")) {
                    DOMConfigurator.configure(configLocation);
                } else {
                    PropertyConfigurator.configure(configLocation);
                }

                configurationDone = true;
                logConfigSource = configLocation;
            }
        } catch (Exception e) {
            logMessages.add("A error occured while configuring the logging framework "
                            + "using the configuration file given by the property "
                            + LOG_CONFIG_PROPERTY_NAME
                            + ". The following exception was thrown: "
                            + e.toString());
        }

        if (!configurationDone) {
            // use default configuration files in the config directory 
            // of the renew distribution for basic configuration properties.
            // Than look for an additional configuration file in the users
            // home directory. Set properties in the users configuration
            // file overwrites the properties set in the default configuration.
            try {
                // find renew distribution home directory
                URL url = getLoaderLocation();
                String base = url.toExternalForm();
                base = base.substring(0, base.lastIndexOf("/"));


                // check if log.home is set. If not set to default location.
                // default location is a "renewlogs" directory in the users home directory
                if (System.getProperty(LOG_HOME_PROPERTY_NAME) == null) {
                    File logDir = new File(System.getProperty("user.home")
                                           + File.separator
                                           + LOG_HOME_DEFAULT_LOCATION);

                    if (!logDir.exists()) {
                        logDir.mkdirs();
                    }
                    if (!logDir.isDirectory()) {
                        // use the parent directory as log files location
                        logDir = logDir.getParentFile();
                    } else {
                        System.setProperty(LOG_HOME_PROPERTY_NAME,
                                           logDir.getAbsolutePath());
                    }
                }

                // use default configuration files to configure the log4j
                // logging framework. First look for log4j.xml.
                url = new URL(base + "/config/" + LOG4J_XML_FILE);
                File defaultConfigFile = new File(new URI(url.toExternalForm()));


                // look for additional configuration file in the user home directory
                // TODO: How should we handle the merging of two xml configuration files.
                //       Currenty only the users configuration file is used to configure
                //       the logging framework if one exists.
                File userConfigFile = new File(System.getProperty("user.home")
                                               + File.separator + "."
                                               + LOG4J_XML_FILE);
                if (userConfigFile.exists() && userConfigFile.isFile()) {
                    // the user config file overrides the default config file.
                    defaultConfigFile = userConfigFile;
                }
                if (defaultConfigFile.exists() && defaultConfigFile.isFile()) {
                    try {
                        DOMConfigurator.configure(defaultConfigFile
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   .getAbsolutePath());

                        configurationDone = true;
                        logConfigSource = defaultConfigFile.getAbsolutePath();
                    } catch (Exception e) {
                        logMessages.add("An error occured during the configuration of "
                                        + "the logging framework using the xml configuration "
                                        + "file in the default config directory of the renew distribution."
                                        + "Error was: " + e.toString());
                    }
                }

                if (!configurationDone) {
                    // try to use log4j.properties file for configuration
                    url = new URL(base + "/config/" + LOG4J_PROPERTIES_FILE);
                    defaultConfigFile = new File(new URI(url.toExternalForm()));
                    userConfigFile = new File(System.getProperty("user.home")
                                              + File.separator + "."
                                              + LOG4J_PROPERTIES_FILE);
                    StringBuffer propsSource = new StringBuffer();
                    Properties defaultProps = new Properties();
                    boolean foundProps = false;
                    if (defaultConfigFile.exists()
                                && defaultConfigFile.isFile()) {
                        defaultProps.load(new FileInputStream(defaultConfigFile));
                        propsSource.append(defaultConfigFile.getAbsolutePath());
                        foundProps = true;
                    }
                    if (userConfigFile.exists() && userConfigFile.isFile()) {
                        defaultProps.load(new FileInputStream(userConfigFile));
                        if (foundProps) {
                            propsSource.append(" and ");
                        }
                        propsSource.append(userConfigFile.getAbsolutePath());
                        foundProps = true;
                    }
                    if (foundProps) {
                        try {
                            PropertyConfigurator.configure(defaultProps);

                            configurationDone = true;
                            logConfigSource = propsSource.toString();
                        } catch (Exception e) {
                            logMessages.add("An error occured during the configuration of "
                                            + "the logging framework using the properties configuration "
                                            + "file(s) "
                                            + propsSource.toString()
                                            + ". Error was: " + e.toString());
                        }
                    }
                }
            } catch (Exception e) {
                logMessages.add(e.toString());
            }

            if (!configurationDone) {
                ConsoleAppender failsafeAppender = new ConsoleAppender(new SimpleLayout());
                failsafeAppender.setThreshold(Level.INFO);
                BasicConfigurator.configure(failsafeAppender);
                logMessages.add("No log4j configuration file found, falling back to failsaife settings.");
                logConfigSource = "failsafe settings";
            }
        }


        // now the log4j framework should be in any way configured,
        // so we can now output any stored log messages.
        if (!logMessages.isEmpty()) {
            Iterator<String> i = logMessages.iterator();
            while (i.hasNext()) {
                logger.error(i.next());
            }
        }
        if (logger.isDebugEnabled()) {
            logger.debug("Used " + logConfigSource + " to configure Log4j.");
        }
    }

    private void initLocationFinders() {
        URL url = getLoaderLocation(); //getClass().getProtectionDomain().getCodeSource().getLocation();
        try {
            url = new URL(url, "plugins/");
        } catch (MalformedURLException e) {
            logger.error("Could not deduce plugins directory near plugin loader: "
                         + e);
            try {
                url = new URL(new File(System.getProperty("user.dir")).toURL(),
                              "plugins/");
            } catch (MalformedURLException e2) {
                logger.error("Could not deduce plugins directory near current directory: "
                             + e2);
                return;
            }
        }
        _locationFinder.addLocationFinder(new PluginSubDirFinder(url));
        _locationFinder.addLocationFinder(new PluginJarLocationFinder(url));
        Iterator<URL> fromFile = getLocations();
        while (fromFile.hasNext()) {
            url = fromFile.next();
            _locationFinder.addLocationFinder(new PluginSubDirFinder(url));
            _locationFinder.addLocationFinder(new PluginJarLocationFinder(url));
        }
    }

    /*
     * return a list of URLs to be added to the location finder
     */
    private Iterator<URL> getLocations() {
        Vector<URL> result = new Vector<URL>();
        Properties userProps = PluginProperties.getUserProperties();
        if (userProps.getProperty(PLUGIN_LOCATIONS_PROPERTY) == null) {
            logger.info("no additional plugin locations set.");
        }
        Collection locations = PropertyHelper.parsePathListString(userProps
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     .getProperty(PLUGIN_LOCATIONS_PROPERTY,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ""));
        Iterator files = locations.iterator();
        while (files.hasNext()) {
            String file = (String) files.next();
            logger.debug("location: " + file);
            URL url = createURLfromString(file);
            logger.debug("as URL: " + url);

            result.add(url);
        }
        return result.iterator();
    }

    private URL createURLfromString(String str) {
        URL url = null;

        // first check if str is a filename of an existing file
        File file = new File(str);
        if (file.exists()) {
            try {
                url = file.toURI().toURL();
                return url;
            } catch (MalformedURLException e) {
                logger.error(e.getMessage(), e);
            }
        }

        // then, try to create a url brute force
        try {
            url = new URL(str);
        } catch (MalformedURLException e) {
            logger.warn("Neither file nor url: " + str);
        }
        return url;
    }

    /**
     * This command prints a list of all available commands.
     */
    public class HelpCommand implements CLCommand {
        /*
        * print all available commands
        */
        public void execute(String[] args, PrintStream response) {
            response.println("available commands:");
            Iterator<String> keys = _commands.keySet().iterator();
            while (keys.hasNext()) {
                String key = keys.next();
                CLCommand command = _commands.get(key);
                response.println(key + "\t-\t" + command.getDescription());
            }
        }

        public String getDescription() {
            return "print a list of all commands";
        }
    }

    /**
     * This class represents a timeout when trying to clean up
     * a plugin.
     * An object must be given to synchronize on; this will
     * be notified when the cleanup worked.
     */
    private static class SynchronizedThread extends Thread {
        private boolean _success = false;
        private boolean _finished = false;
        private Object _toNotify;
        private Command _toExecute;

        public SynchronizedThread(Command toExecute, Object toNotify) {
            _toExecute = toExecute;
            _toNotify = toNotify;
        }

        public void run() {
            _success = _toExecute.execute();
            if (logger.isTraceEnabled()) {
                logger.trace("Execution finished: 1) synching on " + _toNotify);
            }
            synchronized (_toNotify) {
                if (logger.isTraceEnabled()) {
                    logger.trace("Execution finished: 2) notifying on "
                                 + _toNotify);
                }
                _finished = true;
                _toNotify.notify();
            }
            if (logger.isTraceEnabled()) {
                logger.trace("Execution finished: 3) done on " + _toNotify);
            }
        }

        public boolean hadSuccess() {
            return _success;
        }

        public boolean didFinish() {
            return _finished;
        }
    }

    private static interface Command {
        public boolean execute();
    }

    public URL[] getLibs() {
        URLClassLoader urlCL;
        try {
            urlCL = (URLClassLoader) getSystemClassLoader();
        } catch (ClassCastException e) {
            logger.warn("Could not extract URL from SystemClassLoader, no URLClassLoader.",
                        e);
            return null;
        }
        return urlCL.getURLs();
    }
}